/*
  Copyright 2003 John Knottenbelt
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "config.h"
#include "controlledconnection.h"
#include "demuxer.h"
#include "utils.h"
#include "stringutil.h"
#include "inputbuffer.h"
#include "outputbuffer.h"
#include "scheduler.h"
#include "connectionmanager.h"

#include <iostream>
#include <cstring>
#include <vector>
#include <unistd.h>
#include <cerrno>
#include <ctime>

ControlledConnection::ControlledConnection(TunerManager *tm, Scheduler *scheduler, 
					   ConnectionManager *cm,
					   int controlFD, int dataFD)
  : Connection(tm, dataFD),
    controlFD(controlFD),
    scheduler(scheduler),
    cm(cm)
{
  controlInput = new InputBuffer(controlFD, 1024, 4096);
  controlOutput = new OutputBuffer(controlFD, 4096, 16384);

  controlOutput->write("Welcome to dvbd\n");
}

ControlledConnection::~ControlledConnection()
{
  delete controlOutput;
  delete controlInput;
  close(controlFD);
}

void ControlledConnection::writeControl(const std::string &str)
{
  if (!controlOutput->write(str.data(), str.size())) {
      std::cerr << "Client not reading data connection" << std::endl;      
      setRemoveMe(true);
  }
}

void ControlledConnection::notifyUnsubscribe(Source *s)
{
  Connection::notifyUnsubscribe(s);
  if (numDemuxers() && !isRetuning())
    writeControl( "INFO LOST SUBSCRIPTION\n");
}

void ControlledConnection::addSelectFDs(Select &s) const
{
  Connection::addSelectFDs(s);

  // Write data out before reading in more data
  if (controlOutput->getUsed() > 0) 
    controlOutput->addSelectFDs(s);
  else
    controlInput->addSelectFDs(s);
}

bool ControlledConnection::isReady(const Select &s) const
{
  return (Connection::isReady(s) || 
	  controlOutput->isReady(s) || 
	  controlInput->isReady(s));
}

std::string ControlledConnection::getType() const
{
  return "ControlledConnection";
}

bool ControlledConnection::processReady(const Select &s)
{
  if (!Connection::processReady(s))
    return false;

  if (controlOutput && controlOutput->isReady(s)) {
    if (!controlOutput->process())
      return false;
  }
  else if (controlInput && controlInput->isReady(s)) {
    if (!controlInput->process()) 
      return false;

    std::string line;
    if (!controlInput->readLine(line)) {
      if (controlInput->getFree() == 0)
	return false;		// Buffer is full, should have had a
				// line by now
      else
	return true;		// Wait for line to come in
    }

    line = stripTrailingWS(line);

    std::vector<std::string> words;
    splitString(line, ' ', words);
  
    // PRIORITY <a number>
    // (Used to raise or lower the priority of an existing connection)
    if (words[0] == "PRIORITY") {
      if (words.size() != 2) 
	writeControl("ERROR PRIORITY WRONG NUMBER OF ARGS\n");
      else {
	setPriority(toInt(words[1]));
	writeControl("OK\n");
      }
    }
    // LIST <type> <priority>
    else if (words[0] == "LIST") {
      if (words.size() != 3)
	writeControl( "ERROR LIST WRONG NUMBER OF ARGS\n");
      else {
	const std::string &type = words[1];
	int priority = toInt(words[2]);
	StringList channels;

	SavePriority save(this, priority - 1);

	getTunableChannels(type, priority, channels);
	for (StringList::iterator i = channels.begin(); i != channels.end(); i++)
	  writeControl( " " + *i + "\n");
	writeControl( "OK\n");
      }
    }
    // TUNE <type> <channel> <priority>
    // (Used to tune/retune)
    else if (words[0] == "TUNE") {
      if (words.size() != 5) 
	writeControl("ERROR TUNE WRONG NUMBER OF ARGS (Expected 4, got " +
		     ::toString(words.size() - 1) + ")\n");
      else {
	const std::string &type = words[1];
	std::string channel = unescapeWS(words[2]);
	bool convertToPS = toBool(words[3]);
	int priority = toInt(words[4]);

	// Check the arguments
	if (!isValidType(type)) 
	  writeControl( "ERROR TUNE INVALID TYPE\n");
	else if (!isValidChannel(type,channel))
	  writeControl( "ERROR TUNE INVALID CHANNEL ("+channel+")\n");
	else if (!canTune(type, channel, priority)) {
	  writeControl( "ERROR TUNE INSUFFICIENT PRIORITY ("+::toString(priority)+")\n");
	}
	else {
	  bool result = tune(type, channel, convertToPS, priority);
	  if (result)
	    writeControl( "OK\n");
	  else 
	    writeControl( "ERROR TUNE FAILED\n");
	}
      }
    }
    // SCHEDULE <date> <time> <type> <channel> <priority> <duration> <path>
    else if (words[0] == "SCHEDULE") {
      if (words.size() != 10) 
	writeControl("ERROR SCHEDULE WRONG NUMBER OF ARGS (Expected 9, got " +
		     ::toString(words.size() - 1) + ")\n");
      else {
	std::string dateTime = words[1] + " " + words[2];

	struct tm tm;
	if (strptime(dateTime.c_str(), "%Y-%m-%d %H:%M:%S", &tm) == NULL) 
	  writeControl("ERROR SCHEDULE INVALID DATE/TIME\n");
	else {
	  time_t theTime = mktime(&tm);

	  const std::string &type = words[3];
	  const std::string channel = unescapeWS(words[4]);
	  bool convertToPS = toBool(words[5]);
	  int priority = toInt(words[6]);
	  int duration = toInt(words[7]);	// duration in seconds
	  const std::string path = unescapeWS(words[8]);
	  std::string extra;
	  pid_t pid = 0;
	  uid_t uid = 0;
	  gid_t gid = 0;

	  if (words[9].size() >= 2)
	    extra = unescapeWS(words[9].substr(1, words[9].size()-2));
	  else
	    extra = "";

	  if (path.empty() || path[0] != '/')
	    writeControl("ERROR SCHEDULE MUST BE ABSOLUTE PATH\n");
	  else if (!readCredentials(controlFD, pid, uid, gid)) 
	    writeControl("ERROR SCHEDULE FAILED TO READ CREDENTIALS\n");
	  else if (pid == 0)
	    writeControl("ERROR SCHEDULE FAILED TO READ CREDENTIALS\n");
	  else if (!accessOk(path, uid)) 
	    writeControl("ERROR SCHEDULE PATH ACCESS ERROR\n");
	  else if (theTime + duration < getCurrentTime()) 
	    writeControl("ERROR SCHEDULE START TIME IN THE PAST\n");
	  else {
	    std::cerr << "Scheduling recording for " << theTime 
		      << " (now " << time(NULL) << ")\n";

	    Recording r(type, channel, convertToPS, priority, theTime, 
			duration, path, uid, extra);
	    scheduler->add(r);
	    std::cerr << "Job number is " << r.getJob() << "\n";
	    writeControl("JOB " + ::toString(r.getJob()) + "\n");
	    writeControl("OK\n");
	  }
	}
      }
    }
    else if (words[0] == "CHECK") {
      StringList comments;
      scheduler->checkOverlaps(comments);
      for (StringList::iterator i = comments.begin(); i != comments.end(); i++)
	writeControl(" " + *i + "\n");
      writeControl("OK\n");
    }
    else if (words[0] == "REMOVEJOB") {
      if (words.size() != 2) 
	writeControl("ERROR REMOVEJOB WRONG NUMBER OF ARGS (Expected 1, got " +
		     ::toString(words.size() - 1) + ")\n");
      else {
	int job = toInt(words[1]);
	if (!scheduler->remove(job) && !cm->removeJob(job))
	  writeControl("ERROR REMOVEJOB NO SUCH JOB\n");
	else 
	  writeControl("OK\n");
      }
    }
    else if (words[0] == "LISTSCHEDULE") {
      StringList s;
      scheduler->getSchedule(s);
      for (StringList::iterator i = s.begin(); i != s.end(); i++)
	writeControl(" " + *i + "\n");
      writeControl("OK\n");
    }
    else if (words[0] == "NEXTJOBTIME") {
      // Return the time for the start of the next job.
      time_t nextJobTime;
      if (scheduler->getNextJobTime(nextJobTime)) {
	char data[256];
	strftime( data, 256, "%Y-%m-%d %H:%M:%S\n", localtime(&nextJobTime));
	writeControl(data);
	writeControl("OK\n");
      }
      else 
	writeControl("ERROR NEXTJOBTIME NO NEXT JOB\n");
    }
    else if (words[0] == "LISTCONNECTIONS") {
      // List the current connections and their
      // priorities (possibly also their user ids)
      // Note -- It would be great to be able
      // to associate a job number with a connection
      // (in the event that it is indeed a scheduled
      // connection). Is this a hint that perhaps
      // we need a subclass of connection for 
      // scheduled connections?
      
      for (ConnectionManager::const_iterator i = cm->begin(); i != cm->end(); i++) 
	writeControl((*i)->toString() + "\n");
      writeControl("OK\n");
    }
    // OTHER COMMAND
    else {
      writeControl( "ERROR UNRECOGNISED COMMAND\n");
    }

    if (controlOutput->hasOverflowed()) {
      std::cerr << "Control output buffer overflowed. Closing connection." << std::endl;
      return false;
    }
  }

  return true;
}  
