/*
  Copyright 2003 John Knottenbelt
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "config.h"
#include "utils.h"
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream>
#include <cctype>
#include <cerrno>

void setNonBlocking(int fd)
{
  long flags = fcntl(fd, F_GETFL, 0); 
  flags |= O_NONBLOCK;
  fcntl(fd, F_SETFL, flags);
}

int readLineStripTrailingWS(int fd, std::string &line)
{
  char data[1024];
  int bytesread = read(fd, data, 1023);

  if (bytesread <= 0)
    return bytesread;

  data[bytesread] = '\x0';

  // Strip trailing whitespace;
  for (int j = bytesread - 1; j >= 0 && isspace(data[j]); j--)
    data[j] = '\x0';

  line = data;
  return bytesread;
}

void writeString(int fd, const std::string &line)
{
  write(fd, line.data(), line.size());
}

std::string escapeWS(const std::string &channel)
{
  std::string result;

  for (unsigned i = 0; i < channel.size(); i++) {
    switch (channel[i]) {
    case '\\':
      result += "\\\\";
      break;
    case ' ':
      result += "\\_";
      break;
    default:
      result += channel[i];
    }
  }

  return result;
}

std::string unescapeWS(const std::string &channel)
{
  std::string result;
  bool escaped = false;

  for (unsigned i = 0; i < channel.size(); i++) {
    if (channel[i] == '\\')
      escaped = true;
    else {
      char ch = channel[i];

      if (escaped && ch == '_') {
	ch = ' ';
	escaped = false;
      }

      result += ch;
    }
  }
  return result;
}

bool readCredentials(int fd, pid_t &pid, uid_t &uid, gid_t &gid)
{
  int value = 1;
  if (setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &value, sizeof(value)) < 0) 
    std::cerr << "Failed to set SOL_SOCKET SO_PASSCRED: " << strerror(errno) << "\n";

  struct ucred ucred;
  socklen_t size = sizeof(ucred);
  if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &ucred, &size) < 0) {
    std::cerr << "Failed to get credentials from UNIX socket: " 
	      << strerror(errno) << "\n";
    return false;  
  }

  pid = ucred.pid;
  uid = ucred.uid;
  gid = ucred.gid;
  return true;
}

bool accessOk(const std::string &path, uid_t uid)
{
  std::string::size_type lastSlash = path.find_last_of('/');
  if (lastSlash == std::string::npos) 
    return false;

  std::string dirPart = path.substr(0, lastSlash);
  std::string filePart = path.substr(lastSlash + 1);

  // Client must own directory
  struct stat info;
  if (stat(dirPart.c_str(), &info) != 0)
    return false;
  
  if (info.st_uid != uid)
    return false;

  // But must still be writable by this process
  if (access(path.c_str(), F_OK) == 0)
    return access(path.c_str(), W_OK) == 0;
  else 
    return access(dirPart.c_str(), W_OK) == 0;
}

